# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import ctypes
import inspect
import os
import struct
import sys

import six

# Each edit operation is assigned different cost, such as:
#  'w' means swap operation, the cost is 0;
#  's' means substitution operation, the cost is 2;
#  'a' means insertion operation, the cost is 1;
#  'd' means deletion operation, the cost is 3;
# The smaller cost results in the better similarity.
COST = {'w': 0, 's': 2, 'a': 1, 'd': 3}


if hasattr(inspect, 'getfullargspec'):
    getargspec = inspect.getfullargspec
else:
    getargspec = inspect.getargspec


def damerau_levenshtein(s1, s2, cost):
    """Calculates the Damerau-Levenshtein distance between two strings.

    The Levenshtein distance says the minimum number of single-character edits
    (i.e. insertions, deletions, swap or substitution) required to change one
    string to the other.
    The idea is to reserve a matrix to hold the Levenshtein distances between
    all prefixes of the first string and all prefixes of the second, then we
    can compute the values in the matrix in a dynamic programming fashion. To
    avoid a large space complexity, only the last three rows in the matrix is
    needed.(row2 holds the current row, row1 holds the previous row, and row0
    the row before that.)

    More details:
        https://en.wikipedia.org/wiki/Levenshtein_distance
        https://github.com/git/git/commit/8af84dadb142f7321ff0ce8690385e99da8ede2f
    """

    if s1 == s2:
        return 0

    len1 = len(s1)
    len2 = len(s2)

    if len1 == 0:
        return len2 * cost['a']
    if len2 == 0:
        return len1 * cost['d']

    row1 = [i * cost['a'] for i in range(len2 + 1)]
    row2 = row1[:]
    row0 = row1[:]

    for i in range(len1):
        row2[0] = (i + 1) * cost['d']

        for j in range(len2):

            # substitution
            sub_cost = row1[j] + (s1[i] != s2[j]) * cost['s']

            # insertion
            ins_cost = row2[j] + cost['a']

            # deletion
            del_cost = row1[j + 1] + cost['d']

            # swap
            swp_condition = ((i > 0) and
                             (j > 0) and
                             (s1[i - 1] == s2[j]) and
                             (s1[i] == s2[j - 1])
                             )

            # min cost
            if swp_condition:
                swp_cost = row0[j - 1] + cost['w']
                p_cost = min(sub_cost, ins_cost, del_cost, swp_cost)
            else:
                p_cost = min(sub_cost, ins_cost, del_cost)

            row2[j + 1] = p_cost

        row0, row1, row2 = row1, row2, row0

    return row1[-1]


def terminal_width(stdout):
    if hasattr(os, 'get_terminal_size'):
        # python 3.3 onwards has built-in support for getting terminal size
        try:
            return os.get_terminal_size().columns
        except OSError:
            return None

    if sys.platform == 'win32':
        return _get_terminal_width_windows(stdout)
    else:
        return _get_terminal_width_ioctl(stdout)


def _get_terminal_width_windows(stdout):
    STD_INPUT_HANDLE = -10
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12

    std_to_win_handle = {
        sys.stdin: STD_INPUT_HANDLE,
        sys.stdout: STD_OUTPUT_HANDLE,
        sys.stderr: STD_ERROR_HANDLE}

    std_handle = std_to_win_handle.get(stdout)
    if not std_handle:
        return None

    handle = ctypes.windll.kernel32.GetStdHandle(std_handle)
    csbi = ctypes.create_string_buffer(22)

    res = ctypes.windll.kernel32.GetConsoleScreenBufferInfo(handle, csbi)
    if res:
        (size_x, size_y, cur_pos_x, cur_pos_y, attr,
         left, top, right, bottom, max_size_x, max_size_y) = struct.unpack(
            "hhhhHhhhhhh", csbi.raw)
        return size_x


def _get_terminal_width_ioctl(stdout):
    from fcntl import ioctl
    import termios

    try:
        # winsize structure has 4 unsigned short fields
        winsize = b'\0' * struct.calcsize('hhhh')
        try:
            winsize = ioctl(stdout, termios.TIOCGWINSZ, winsize)
        except IOError:
            return None
        except TypeError:
            # this is raised in unit tests as stdout is sometimes a StringIO
            return None
        winsize = struct.unpack('hhhh', winsize)
        columns = winsize[1]
        if not columns:
            return None
        return columns
    except IOError:
        return None


if six.PY2:
    def getwriter(encoding):
        '''Override codecs.getwriter() to prevent codec errors.

        The StreamWriter returned by codecs.getwriter has an unfortunate
        property, it will attempt to encode every object presented to it's
        write() function. Normally we only want unicode objects to be
        encoded to a byte stream. If bytes are presented (e.g. str in
        Python2) we make the assumption those bytes represent an already
        encoded text stream or they are indeed binary bytes and hence
        should not be encoded.

        When the core StreamWriter attempts to encode a str object Python
        will first promote the str object to a unicode object. The
        promotion of str to unicode requires the str bytes to be
        decoded. However the encoding associated with the str object is
        not known therefore Python applies the default-encoding which is
        ASCII. In the case where the str object contains utf-8 encoded
        non-ASCII characters a decoding error is raised. By not attempting
        to encode a byte stream we avoid this error.

        It really does not make much sense to try and encode a byte
        stream. First of all a byte stream should not be encoded if it's
        not text (e.g. binary data). If the byte stream is encoded text
        the only way to re-encode it is if we known it's encoding so we
        can decode it into a canonical form (e.g. unicode). Thus to
        re-encode it we encode from the canonical form (e.g. unicode) to
        the new binary encoding. The problem in Python2 is we never know
        if the bytes in a str object are text or binary data and if it's
        text which encoding it is, hence we should not try to apply
        an encoding to a str object.
        '''
        class _StreamWriter(codecs.StreamWriter):
            def __init__(self, stream, errors='strict'):
                codecs.StreamWriter.__init__(self, stream, errors)

            def encode(self, msg, errors='strict'):
                if isinstance(msg, six.text_type):
                    return self.encoder(msg, errors)
                return msg, len(msg)

        _StreamWriter.encoder = codecs.getencoder(encoding)
        _StreamWriter.encoding = encoding
        return _StreamWriter

else:
    getwriter = codecs.getwriter



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class for providing a list of data as output.
"""
import abc
import operator
import six

from . import display


@six.add_metaclass(abc.ABCMeta)
class Lister(display.DisplayCommandBase):
    """Command base class for providing a list of data as output.
    """

    @property
    def formatter_namespace(self):
        return 'cliff.formatter.list'

    @property
    def formatter_default(self):
        return 'table'

    @property
    def need_sort_by_cliff(self):
        """Whether sort procedure is performed by cliff itself.

        Should be overridden (return False) when there is a need to implement
        custom sorting procedure or data is already sorted."""
        return True

    @abc.abstractmethod
    def take_action(self, parsed_args):
        """Return a tuple containing the column names and an iterable
        containing the data to be listed.
        """

    def get_parser(self, prog_name):
        parser = super(Lister, self).get_parser(prog_name)
        group = self._formatter_group
        group.add_argument(
            '--sort-column',
            action='append',
            default=[],
            dest='sort_columns',
            metavar='SORT_COLUMN',
            help=("specify the column(s) to sort the data (columns specified "
                  "first have a priority, non-existing columns are ignored), "
                  "can be repeated")
        )
        return parser

    def produce_output(self, parsed_args, column_names, data):
        if parsed_args.sort_columns and self.need_sort_by_cliff:
            indexes = [column_names.index(c) for c in parsed_args.sort_columns
                       if c in column_names]
            if indexes:
                data = sorted(data, key=operator.itemgetter(*indexes))
        (columns_to_include, selector) = self._generate_columns_and_selector(
            parsed_args, column_names)
        if selector:
            # Generator expression to only return the parts of a row
            # of data that the user has expressed interest in
            # seeing. We have to convert the compress() output to a
            # list so the table formatter can ask for its length.
            data = (list(self._compress_iterable(row, selector))
                    for row in data)
        self.formatter.emit_list(columns_to_include,
                                 data,
                                 self.app.stdout,
                                 parsed_args,
                                 )
        return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import unittest

from cliff import _argparse


class TestArgparse(unittest.TestCase):

    def test_argument_parser(self):
        _argparse.ArgumentParser(conflict_handler='ignore')

    def test_argument_parser_add_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        parser.add_argument_group()

    def test_argument_parser_add_mutually_exclusive_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        parser.add_mutually_exclusive_group()

    def test_argument_parser_add_nested_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        group = parser.add_argument_group()
        group.add_argument_group()

    def test_argument_parser_add_nested_mutually_exclusive_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        group = parser.add_argument_group()
        group.add_mutually_exclusive_group()

    def test_argument_parser_add_mx_nested_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        group = parser.add_mutually_exclusive_group()
        group.add_argument_group()

    def test_argument_parser_add_mx_nested_mutually_exclusive_group(self):
        parser = _argparse.ArgumentParser(conflict_handler='ignore')
        group = parser.add_mutually_exclusive_group()
        group.add_mutually_exclusive_group()



# Copyright (C) 2017, Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import textwrap

from cliff import sphinxext
from cliff.tests import base


class TestSphinxExtension(base.TestBase):

    def test_empty_help(self):
        """Handle positional and optional actions without help messages."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('name', action='store')
        parser.add_argument('--language', dest='lang')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world [--language LANG] name

        .. option:: --language <LANG>

        .. option:: name
        """).lstrip(), output)

    def test_nonempty_help(self):
        """Handle positional and optional actions with help messages."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('name', help='user name')
        parser.add_argument('--language', dest='lang',
                            help='greeting language')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world [--language LANG] name

        .. option:: --language <LANG>

            greeting language

        .. option:: name

            user name
        """).lstrip(), output)

    def test_description_epilog(self):
        """Handle a parser description, epilog."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False,
                                         description='A "Hello, World" app.',
                                         epilog='What am I doing down here?')
        parser.add_argument('name', action='store')
        parser.add_argument('--language', dest='lang')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        A "Hello, World" app.

        .. program:: hello-world
        .. code-block:: shell

            hello-world [--language LANG] name

        .. option:: --language <LANG>

        .. option:: name

        What am I doing down here?
        """).lstrip(), output)

    def test_flag(self):
        """Handle a boolean argparse action."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('name', help='user name')
        parser.add_argument('--translate', action='store_true',
                            help='translate to local language')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world [--translate] name

        .. option:: --translate

            translate to local language

        .. option:: name

            user name
        """).lstrip(), output)

    def test_supressed(self):
        """Handle a supressed action."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('name', help='user name')
        parser.add_argument('--variable', help=argparse.SUPPRESS)

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world name


        .. option:: name

            user name
        """).lstrip(), output)

    def test_metavar(self):
        """Handle an option with a metavar."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('names', metavar='<NAME>', nargs='+',
                            help='a user name')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world <NAME> [<NAME> ...]

        .. option:: NAME

            a user name
        """).lstrip(), output)

    def test_multiple_opts(self):
        """Correctly output multiple opts on separate lines."""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('name', help='user name')
        parser.add_argument('--language', dest='lang',
                            help='greeting language')
        parser.add_argument('--translate', action='store_true',
                            help='translate to local language')
        parser.add_argument('--write-to-var-log-something-or-other',
                            action='store_true',
                            help='a long opt to force wrapping')
        parser.add_argument('--required-arg', dest='stuff', required=True,
                            help='a required argument')
        style_group = parser.add_mutually_exclusive_group(required=True)
        style_group.add_argument('--polite', action='store_true',
                                 help='use a polite greeting')
        style_group.add_argument('--profane', action='store_true',
                                 help='use a less polite greeting')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world
                [--language LANG]
                [--translate]
                [--write-to-var-log-something-or-other]
                --required-arg STUFF
                (--polite | --profane)
                name

        .. option:: --language <LANG>

            greeting language

        .. option:: --translate

            translate to local language

        .. option:: --write-to-var-log-something-or-other

            a long opt to force wrapping

        .. option:: --required-arg <STUFF>

            a required argument

        .. option:: --polite

            use a polite greeting

        .. option:: --profane

            use a less polite greeting

        .. option:: name

            user name
        """).lstrip(), output)

    def test_various_option_names_with_hyphen(self):
        """Handle options whose name and/or metavar contain hyphen(s)"""
        parser = argparse.ArgumentParser(prog='hello-world', add_help=False)
        parser.add_argument('--foo-bar', metavar='<foo-bar>',
                            help='foo bar', required=True)
        parser.add_argument('--foo-bar-baz', metavar='<foo-bar-baz>',
                            help='foo bar baz', required=True)
        parser.add_argument('--foo', metavar='<foo>',
                            help='foo', required=True)
        parser.add_argument('--alpha', metavar='<A>',
                            help='alpha')
        parser.add_argument('--alpha-beta', metavar='<A-B>',
                            help='alpha beta')
        parser.add_argument('--alpha-beta-gamma', metavar='<A-B-C>',
                            help='alpha beta gamma')

        output = '\n'.join(sphinxext._format_parser(parser))
        self.assertEqual(textwrap.dedent("""
        .. program:: hello-world
        .. code-block:: shell

            hello-world
                --foo-bar <foo-bar>
                --foo-bar-baz <foo-bar-baz>
                --foo <foo>
                [--alpha <A>]
                [--alpha-beta <A-B>]
                [--alpha-beta-gamma <A-B-C>]

        .. option:: --foo-bar <foo-bar>

            foo bar

        .. option:: --foo-bar-baz <foo-bar-baz>

            foo bar baz

        .. option:: --foo <foo>

            foo

        .. option:: --alpha <A>

            alpha

        .. option:: --alpha-beta <A-B>

            alpha beta

        .. option:: --alpha-beta-gamma <A-B-C>

            alpha beta gamma
        """).lstrip(), output)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters using csv format.
"""

import os
import sys

from .base import ListFormatter
from cliff import columns

import six

if sys.version_info[0] == 3:
    import csv
else:
    import unicodecsv as csv


class CSVLister(ListFormatter):

    QUOTE_MODES = {
        'all': csv.QUOTE_ALL,
        'minimal': csv.QUOTE_MINIMAL,
        'nonnumeric': csv.QUOTE_NONNUMERIC,
        'none': csv.QUOTE_NONE,
    }

    def add_argument_group(self, parser):
        group = parser.add_argument_group('CSV Formatter')
        group.add_argument(
            '--quote',
            choices=sorted(self.QUOTE_MODES.keys()),
            dest='quote_mode',
            default='nonnumeric',
            help='when to include quotes, defaults to nonnumeric',
        )

    def emit_list(self, column_names, data, stdout, parsed_args):
        writer_kwargs = dict(
            quoting=self.QUOTE_MODES[parsed_args.quote_mode],
            lineterminator=os.linesep,
            escapechar='\\',
        )

        # In Py2 we replace the csv module with unicodecsv because the
        # Py2 csv module cannot handle unicode. unicodecsv encodes
        # unicode objects based on the value of it's encoding keyword
        # with the result unicodecsv emits encoded bytes in a str
        # object. The utils.getwriter assures no attempt is made to
        # re-encode the encoded bytes in the str object.

        if six.PY2:
            writer_kwargs['encoding'] = (getattr(stdout, 'encoding', None)
                                         or 'utf-8')

        writer = csv.writer(stdout, **writer_kwargs)
        writer.writerow(column_names)
        for row in data:
            writer.writerow(
                [(six.text_type(c.machine_readable())
                  if isinstance(c, columns.FormattableColumn)
                  else c)
                 for c in row]
            )
        return



import logging

from cliff.command import Command


class Simple(Command):
    "A simple command that prints a message."

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        self.log.info('sending greeting')
        self.log.debug('debugging')
        self.app.stdout.write('hi!\n')


class Error(Command):
    "Always raises an error"

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        self.log.info('causing error')
        raise RuntimeError('this is the expected exception')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Discover and lookup command plugins.
"""

import logging

import pkg_resources

from . import utils


LOG = logging.getLogger(__name__)


def _get_commands_by_partial_name(args, commands):
    n = len(args)
    candidates = []
    for command_name in commands:
        command_parts = command_name.split()
        if len(command_parts) != n:
            continue
        if all(command_parts[i].startswith(args[i]) for i in range(n)):
            candidates.append(command_name)
    return candidates


class EntryPointWrapper(object):
    """Wrap up a command class already imported to make it look like a plugin.
    """

    def __init__(self, name, command_class):
        self.name = name
        self.command_class = command_class

    def load(self, require=False):
        return self.command_class


class CommandManager(object):
    """Discovers commands and handles lookup based on argv data.

    :param namespace: String containing the setuptools entrypoint namespace
                      for the plugins to be loaded. For example,
                      ``'cliff.formatter.list'``.
    :param convert_underscores: Whether cliff should convert underscores to
                                spaces in entry_point commands.
    """
    def __init__(self, namespace, convert_underscores=True):
        self.commands = {}
        self._legacy = {}
        self.namespace = namespace
        self.convert_underscores = convert_underscores
        self._load_commands()

    def _load_commands(self):
        # NOTE(jamielennox): kept for compatibility.
        if self.namespace:
            self.load_commands(self.namespace)

    def load_commands(self, namespace):
        """Load all the commands from an entrypoint"""
        for ep in pkg_resources.iter_entry_points(namespace):
            LOG.debug('found command %r', ep.name)
            cmd_name = (ep.name.replace('_', ' ')
                        if self.convert_underscores
                        else ep.name)
            self.commands[cmd_name] = ep
        return

    def __iter__(self):
        return iter(self.commands.items())

    def add_command(self, name, command_class):
        self.commands[name] = EntryPointWrapper(name, command_class)

    def add_legacy_command(self, old_name, new_name):
        """Map an old command name to the new name.

        :param old_name: The old command name.
        :type old_name: str
        :param new_name: The new command name.
        :type new_name: str

        """
        self._legacy[old_name] = new_name

    def find_command(self, argv):
        """Given an argument list, find a command and
        return the processor and any remaining arguments.
        """
        start = self._get_last_possible_command_index(argv)
        for i in range(start, 0, -1):
            name = ' '.join(argv[:i])
            search_args = argv[i:]
            # The legacy command handling may modify name, so remember
            # the value we actually found in argv so we can return it.
            return_name = name
            # Convert the legacy command name to its new name.
            if name in self._legacy:
                name = self._legacy[name]

            found = None
            if name in self.commands:
                found = name
            else:
                candidates = _get_commands_by_partial_name(
                    argv[:i], self.commands)
                if len(candidates) == 1:
                    found = candidates[0]
            if found:
                cmd_ep = self.commands[found]
                if hasattr(cmd_ep, 'resolve'):
                    cmd_factory = cmd_ep.resolve()
                else:
                    # NOTE(dhellmann): Some fake classes don't take
                    # require as an argument. Yay?
                    arg_spec = utils.getargspec(cmd_ep.load)
                    if 'require' in arg_spec[0]:
                        cmd_factory = cmd_ep.load(require=False)
                    else:
                        cmd_factory = cmd_ep.load()
                return (cmd_factory, return_name, search_args)
        else:
            raise ValueError('Unknown command %r' %
                             (argv,))

    def _get_last_possible_command_index(self, argv):
        """Returns the index after the last argument
        in argv that can be a command word
        """
        for i, arg in enumerate(argv):
            if arg.startswith('-'):
                return i
        return len(argv)
